"""
This contains those builtin functions from SPARQL that may be of interest to spy developers
"""

import datetime as py_datetime  # naming conflict with function within this module
import hashlib
import math
import operator as pyop  # python operators
import random
import re
import sys
import uuid
import warnings
from decimal import ROUND_HALF_UP, Decimal, InvalidOperation
from functools import reduce
from urllib.parse import quote

import isodate
from pyparsing import ParseResults

from rdflib import RDF, XSD, BNode, Literal, URIRef, Variable
from rdflib.plugins.sparql.datatypes import (
    XSD_DateTime_DTs,
    XSD_DTs,
    XSD_Duration_DTs,
    type_promotion,
)
from rdflib.plugins.sparql.parserutils import CompValue, Expr
from rdflib.plugins.sparql.sparql import SPARQLError, SPARQLTypeError
from rdflib.term import Node
from rdflib.plugins.sparql import operators



def isBlank(arg):
    return Literal(isinstance(arg, BNode))


def isLiteral(arg):
    return Literal(isinstance(arg, Literal))


def isIRI(arg):
    return Literal(isinstance(arg, URIRef))


def isNumeric(arg):
    try:
        numeric(arg)
        return Literal(True)
    except:
        return Literal(False)


def abs(arg):
    """
    http://www.w3.org/TR/sparql11-query/#func-abs
    """

    return Literal(abs(numeric(arg)))



def RAND():
    """
    http://www.w3.org/TR/sparql11-query/#idp2133952
    """

    return Literal(random.random())


def UUID():
    """
    http://www.w3.org/TR/sparql11-query/#func-strdt
    """

    return URIRef(uuid.uuid4().urn)


def STRUUID():
    """
    http://www.w3.org/TR/sparql11-query/#func-strdt
    """

    return Literal(str(uuid.uuid4()))


def MD5(arg):
    s = string(arg).encode("utf-8")
    return Literal(hashlib.md5(s).hexdigest())


def SHA1(arg):
    s = string(arg).encode("utf-8")
    return Literal(hashlib.sha1(s).hexdigest())


def SHA256(arg):
    s = string(arg).encode("utf-8")
    return Literal(hashlib.sha256(s).hexdigest())


def SHA384(arg):
    s = string(arg).encode("utf-8")
    return Literal(hashlib.sha384(s).hexdigest())


def SHA512(arg):
    s = string(arg).encode("utf-8")
    return Literal(hashlib.sha512(s).hexdigest())


def COALESCE(*args):
    """
    http://www.w3.org/TR/sparql11-query/#func-coalesce
    """
    for x in args:
        if x is not None and not isinstance(x, (SPARQLError, Variable)):
            return x
    raise SPARQLError("COALESCE got no arguments that did not evaluate to an error")


def ceil(arg):
    """
    http://www.w3.org/TR/sparql11-query/#func-ceil
    """

    l_ = arg
    return Literal(int(math.ceil(numeric(l_))), datatype=l_.datatype)


def floor(arg):
    """
    http://www.w3.org/TR/sparql11-query/#func-floor
    """
    l_ = arg
    return Literal(int(math.floor(numeric(l_))), datatype=l_.datatype)


def round(arg):
    """
    http://www.w3.org/TR/sparql11-query/#func-round
    """

    # This used to be just math.bound
    # but in py3k bound was changed to
    # "round-to-even" behaviour
    # this is an ugly work-around
    l_ = arg
    v = numeric(l_)
    v = int(Decimal(v).quantize(1, ROUND_HALF_UP))
    return Literal(v, datatype=l_.datatype)


def REGEX(text, pattern, flags=None):
    """
    http://www.w3.org/TR/sparql11-query/#func-regex
    Invokes the XPath fn:matches function to match text against a regular
    expression pattern.
    The regular expression language is defined in XQuery 1.0 and XPath 2.0
    Functions and Operators section 7.6.1 Regular Expression Syntax
    """

    text = string(text)
    pattern = string(pattern)

    cFlag = 0
    if flags:
        # Maps XPath REGEX flags (http://www.w3.org/TR/xpath-functions/#flags)
        # to Python's re flags
        flagMap = dict([("i", re.IGNORECASE), ("s", re.DOTALL), ("m", re.MULTILINE)])
        cFlag = reduce(pyop.or_, [flagMap.get(f, 0) for f in flags])

    return Literal(bool(re.search(str(pattern), text, cFlag)))


def REPLACE(arg, pattern, replacement, flags=None):
    expr = {}
    return operators.Builtin_REPLACE(expr, None)
    """
    http://www.w3.org/TR/sparql11-query/#func-substr
    """
    text = string(arg)
    pattern = string(pattern)
    replacement = string(replacement)

    # python uses \1, xpath/sparql uses $1
    replacement = re.sub("\\$([0-9]*)", r"\\\1", replacement)

    def _r(m):

        # Now this is ugly.
        # Python has a "feature" where unmatched groups return None
        # then re.sub chokes on this.
        # see http://bugs.python.org/issue1519638 , fixed and errs in py3.5

        # this works around and hooks into the internal of the re module...

        # the match object is replaced with a wrapper that
        # returns "" instead of None for unmatched groups

        class _m:
            def __init__(self, m):
                self.m = m
                self.string = m.string

            def group(self, n):
                return m.group(n) or ""

        return re._expand(pattern, _m(m), replacement)

    cFlag = 0
    if flags:
        # Maps XPath REGEX flags (http://www.w3.org/TR/xpath-functions/#flags)
        # to Python's re flags
        flagMap = dict([("i", re.IGNORECASE), ("s", re.DOTALL), ("m", re.MULTILINE)])
        cFlag = reduce(pyop.or_, [flagMap.get(f, 0) for f in flags])

        # @@FIXME@@ either datatype OR lang, NOT both

    # this is necessary due to different treatment of unmatched groups in
    # python versions. see comments above in _r(m).
    compat_r = str(replacement) if sys.version_info[:2] >= (3, 5) else _r

    return Literal(
        re.sub(str(pattern), compat_r, text, cFlag),
        datatype=text.datatype,
        lang=text.language,
    )


def STRDT(arg):
    """
    http://www.w3.org/TR/sparql11-query/#func-strdt
    """

    return Literal(str(arg1), datatype=arg2)


def STRLANG(arg):
    """
    http://www.w3.org/TR/sparql11-query/#func-strlang
    """

    s = string(arg1)
    if s.language or s.datatype:
        raise SPARQLError("STRLANG expects a simple literal")

    # TODO: normalisation of lang tag to lower-case
    # should probably happen in literal __init__
    return Literal(str(s), lang=str(arg2).lower())


def CONCAT(arg):
    """
    http://www.w3.org/TR/sparql11-query/#func-concat
    """

    # dt/lang passed on only if they all match

    dt = set(x.datatype for x in arg)
    dt = dt.pop() if len(dt) == 1 else None

    lang = set(x.language for x in arg)
    lang = lang.pop() if len(lang) == 1 else None

    return Literal("".join(string(x) for x in arg), datatype=dt, lang=lang)


def _compatibleStrings(a, b):
    string(a)
    string(b)

    if b.language and a.language != b.language:
        raise SPARQLError("incompatible arguments to str functions")


def STRSTARTS(arg):
    """
    http://www.w3.org/TR/sparql11-query/#func-strstarts
    """

    a = arg1
    b = arg2
    _compatibleStrings(a, b)

    return Literal(a.startswith(b))


def STRENDS(arg):
    """
    http://www.w3.org/TR/sparql11-query/#func-strends
    """
    a = arg1
    b = arg2

    _compatibleStrings(a, b)

    return Literal(a.endswith(b))


def STRBEFORE(arg):
    """
    http://www.w3.org/TR/sparql11-query/#func-strbefore
    """

    a = arg1
    b = arg2
    _compatibleStrings(a, b)

    i = a.find(b)
    if i == -1:
        return Literal("")
    else:
        return Literal(a[:i], lang=a.language, datatype=a.datatype)


def STRAFTER(arg):
    """
    http://www.w3.org/TR/sparql11-query/#func-strafter
    """

    a = arg1
    b = arg2
    _compatibleStrings(a, b)

    i = a.find(b)
    if i == -1:
        return Literal("")
    else:
        return Literal(a[i + len(b) :], lang=a.language, datatype=a.datatype)


def CONTAINS(arg):
    """
    http://www.w3.org/TR/sparql11-query/#func-strcontains
    """

    a = arg1
    b = arg2
    _compatibleStrings(a, b)

    return Literal(b in a)


def ENCODE_FOR_URI(arg):
    return Literal(quote(string(arg).encode("utf-8")))


def SUBSTR(arg):
    """
    http://www.w3.org/TR/sparql11-query/#func-substr
    """

    a = string(arg)

    start = numeric(expr.start) - 1

    length = expr.length
    if length is not None:
        length = numeric(length) + start

    return Literal(a[start:length], lang=a.language, datatype=a.datatype)


def STRLEN(e, ctx):
    l_ = string(e.arg)

    return Literal(len(l_))


def STR(e, ctx):
    arg = e.arg
    if isinstance(arg, SPARQLError):
        raise arg
    return Literal(str(arg))  # plain literal


def LCASE(e, ctx):
    l_ = string(e.arg)

    return Literal(l_.lower(), datatype=l_.datatype, lang=l_.language)


def LANGMATCHES(e, ctx):
    """
    http://www.w3.org/TR/sparql11-query/#func-langMatches
    """
    langTag = string(e.arg1)
    langRange = string(e.arg2)

    if str(langTag) == "":
        return Literal(False)  # nothing matches empty!

    return Literal(_lang_range_check(langRange, langTag))


def NOW(e, ctx):
    """
    http://www.w3.org/TR/sparql11-query/#func-now
    """
    return Literal(ctx.now)


def YEAR(e, ctx):
    d = date(e.arg)
    return Literal(d.year)


def MONTH(e, ctx):
    d = date(e.arg)
    return Literal(d.month)


def DAY(e, ctx):
    d = date(e.arg)
    return Literal(d.day)


def HOURS(e, ctx):
    d = datetime(e.arg)
    return Literal(d.hour)


def MINUTES(e, ctx):
    d = datetime(e.arg)
    return Literal(d.minute)


def SECONDS(e, ctx):
    """
    http://www.w3.org/TR/sparql11-query/#func-seconds
    """
    d = datetime(e.arg)
    return Literal(d.second, datatype=XSD.decimal)


def TIMEZONE(e, ctx):
    """
    http://www.w3.org/TR/sparql11-query/#func-timezone
    :returns: the timezone part of arg as an xsd:dayTimeDuration.
    :raises: an error if there is no timezone.
    """
    dt = datetime(e.arg)
    if not dt.tzinfo:
        raise SPARQLError("datatime has no timezone: %r" % dt)

    delta = dt.utcoffset()

    d = delta.days
    s = delta.seconds
    neg = ""

    if d < 0:
        s = -24 * 60 * 60 * d - s
        d = 0
        neg = "-"

    h = s / (60 * 60)
    m = (s - h * 60 * 60) / 60
    s = s - h * 60 * 60 - m * 60

    tzdelta = "%sP%sT%s%s%s" % (
        neg,
        "%dD" % d if d else "",
        "%dH" % h if h else "",
        "%dM" % m if m else "",
        "%dS" % s if not d and not h and not m else "",
    )

    return Literal(tzdelta, datatype=XSD.dayTimeDuration)


def TZ(e, ctx):
    d = datetime(e.arg)
    if not d.tzinfo:
        return Literal("")
    n = d.tzinfo.tzname(d)
    if n == "UTC":
        n = "Z"
    return Literal(n)


def UCASE(e, ctx):
    l_ = string(e.arg)

    return Literal(l_.upper(), datatype=l_.datatype, lang=l_.language)


def LANG(e, ctx):
    """
    http://www.w3.org/TR/sparql11-query/#func-lang
    Returns the language tag of ltrl, if it has one. It returns "" if ltrl has
    no language tag. Note that the RDF data model does not include literals
    with an empty language tag.
    """

    l_ = literal(e.arg)
    return Literal(l_.language or "")


def DATATYPE(e, ctx):
    l_ = e.arg
    if not isinstance(l_, Literal):
        raise SPARQLError("Can only get datatype of literal: %r" % l_)
    if l_.language:
        return RDF.langString
    if not l_.datatype and not l_.language:
        return XSD.string
    return l_.datatype


def sameTerm(e, ctx):
    a = e.arg1
    b = e.arg2
    return Literal(a == b)


def BOUND(e, ctx):
    """
    http://www.w3.org/TR/sparql11-query/#func-bound
    """
    n = e.get("arg", variables=True)

    return Literal(not isinstance(n, Variable))


def EXISTS(e, ctx):
    # damn...
    from rdflib.plugins.sparql.evaluate import evalPart

    exists = e.name == "EXISTS"

    ctx = ctx.ctx.thaw(ctx)  # hmm
    for x in evalPart(ctx, e.graph):
        return Literal(exists)
    return Literal(not exists)


_CUSTOM_FUNCTIONS = {}


def register_custom_function(uri, func, override=False, raw=False):
    """
    Register a custom SPARQL function.
    By default, the function will be passed the RDF terms in the argument list.
    If raw is True, the function will be passed an Expression and a Context.
    The function must return an RDF term, or raise a SparqlError.
    """
    if not override and uri in _CUSTOM_FUNCTIONS:
        raise ValueError("A function is already registered as %s" % uri.n3())
    _CUSTOM_FUNCTIONS[uri] = (func, raw)


def custom_function(uri, override=False, raw=False):
    """
    Decorator version of :func:`register_custom_function`.
    """

    def decorator(func):
        register_custom_function(uri, func, override=override, raw=raw)
        return func

    return decorator


def unregister_custom_function(uri, func=None):
    """
    The 'func' argument is included for compatibility with existing code.
    A previous implementation checked that the function associated with
    the given uri was actually 'func', but this is not necessary as the
    uri should uniquely identify the function.
    """
    if _CUSTOM_FUNCTIONS.get(uri):
        del _CUSTOM_FUNCTIONS[uri]
    else:
        warnings.warn("This function is not registered as %s" % uri.n3())


def Function(e, ctx):
    """
    Custom functions and casts
    """
    pair = _CUSTOM_FUNCTIONS.get(e.iri)
    if pair is None:
        # no such function is registered
        raise SPARQLError("Unknown function %r" % e.iri)
    func, raw = pair
    if raw:
        # function expects expression and context
        return func(e, ctx)
    else:
        # function expects the argument list
        try:
            return func(*e.expr)
        except TypeError as ex:
            # wrong argument number
            raise SPARQLError(*ex.args)


@custom_function(XSD.string, raw=True)
@custom_function(XSD.dateTime, raw=True)
@custom_function(XSD.float, raw=True)
@custom_function(XSD.double, raw=True)
@custom_function(XSD.decimal, raw=True)
@custom_function(XSD.integer, raw=True)
@custom_function(XSD.boolean, raw=True)
def default_cast(e, ctx):
    if not e.expr:
        raise SPARQLError("Nothing given to cast.")
    if len(e.expr) > 1:
        raise SPARQLError("Cannot cast more than one thing!")

    x = e.expr[0]

    if e.iri == XSD.string:

        if isinstance(x, (URIRef, Literal)):
            return Literal(x, datatype=XSD.string)
        else:
            raise SPARQLError("Cannot cast term %r of type %r" % (x, type(x)))

    if not isinstance(x, Literal):
        raise SPARQLError("Can only cast Literals to non-string data-types")

    if x.datatype and not x.datatype in XSD_DTs:
        raise SPARQLError("Cannot cast literal with unknown datatype: %r" % x.datatype)

    if e.iri == XSD.dateTime:
        if x.datatype and x.datatype not in (XSD.dateTime, XSD.string):
            raise SPARQLError("Cannot cast %r to XSD:dateTime" % x.datatype)
        try:
            return Literal(isodate.parse_datetime(x), datatype=e.iri)
        except:
            raise SPARQLError("Cannot interpret '%r' as datetime" % x)

    if x.datatype == XSD.dateTime:
        raise SPARQLError("Cannot cast XSD.dateTime to %r" % e.iri)

    if e.iri in (XSD.float, XSD.double):
        try:
            return Literal(float(x), datatype=e.iri)
        except:
            raise SPARQLError("Cannot interpret '%r' as float" % x)

    elif e.iri == XSD.decimal:
        if "e" in x or "E" in x:  # SPARQL/XSD does not allow exponents in decimals
            raise SPARQLError("Cannot interpret '%r' as decimal" % x)
        try:
            return Literal(Decimal(x), datatype=e.iri)
        except:
            raise SPARQLError("Cannot interpret '%r' as decimal" % x)

    elif e.iri == XSD.integer:
        try:
            return Literal(int(x), datatype=XSD.integer)
        except:
            raise SPARQLError("Cannot interpret '%r' as int" % x)

    elif e.iri == XSD.boolean:
        # # I would argue that any number is True...
        # try:
        #     return Literal(bool(int(x)), datatype=XSD.boolean)
        # except:
        if x.lower() in ("1", "true"):
            return Literal(True)
        if x.lower() in ("0", "false"):
            return Literal(False)
        raise SPARQLError("Cannot interpret '%r' as bool" % x)


def UnaryNot(arg):
    return Literal(not EBV(expr.expr))


def UnaryMinus(arg):
    return Literal(-numeric(expr.expr))


def UnaryPlus(arg):
    return Literal(+numeric(expr.expr))


def MultiplicativeExpression(e, ctx):

    expr = e.expr
    other = e.other

    # because of the way the mul-expr production handled operator precedence
    # we sometimes have nothing to do
    if other is None:
        return expr
    try:
        res = Decimal(numeric(expr))
        for op, f in zip(e.op, other):
            f = numeric(f)

            if type(f) == float:
                res = float(res)

            if op == "*":
                res *= f
            else:
                res /= f
    except (InvalidOperation, ZeroDivisionError):
        raise SPARQLError("divide by 0")

    return Literal(res)


def AdditiveExpression(e, ctx):

    expr = e.expr
    other = e.other

    # because of the way the add-expr production handled operator precedence
    # we sometimes have nothing to do
    if other is None:
        return expr

    # handling arithmetic(addition/subtraction) of dateTime, date, time
    # and duration datatypes (if any)
    if hasattr(expr, "datatype") and (
        expr.datatype in XSD_DateTime_DTs or expr.datatype in XSD_Duration_DTs
    ):

        res = dateTimeObjects(expr)
        dt = expr.datatype

        for op, term in zip(e.op, other):

            # check if operation is datetime,date,time operation over
            # another datetime,date,time datatype
            if dt in XSD_DateTime_DTs and dt == term.datatype and op == "-":
                # checking if there are more than one datetime operands -
                # in that case it doesn't make sense for example
                # ( dateTime1 - dateTime2 - dateTime3 ) is an invalid operation
                if len(other) > 1:
                    error_message = "Can't evaluate multiple %r arguments"
                    raise SPARQLError(error_message, dt.datatype)
                else:
                    n = dateTimeObjects(term)
                    res = calculateDuration(res, n)
                    return res

            # datetime,date,time +/- duration,dayTimeDuration,yearMonthDuration
            elif dt in XSD_DateTime_DTs and term.datatype in XSD_Duration_DTs:
                n = dateTimeObjects(term)
                res = calculateFinalDateTime(res, dt, n, term.datatype, op)
                return res

            # duration,dayTimeDuration,yearMonthDuration + datetime,date,time
            elif dt in XSD_Duration_DTs and term.datatype in XSD_DateTime_DTs:
                if op == "+":
                    n = dateTimeObjects(term)
                    res = calculateFinalDateTime(res, dt, n, term.datatype, op)
                    return res

            # rest are invalid types
            else:
                raise SPARQLError("Invalid DateTime Operations")

    # handling arithmetic(addition/subtraction) of numeric datatypes (if any)
    else:
        res = numeric(expr)

        dt = expr.datatype

        for op, term in zip(e.op, other):
            n = numeric(term)
            if isinstance(n, Decimal) and isinstance(res, float):
                n = float(n)
            if isinstance(n, float) and isinstance(res, Decimal):
                res = float(res)

            dt = type_promotion(dt, term.datatype)

            if op == "+":
                res += n
            else:
                res -= n

        return Literal(res, datatype=dt)


def RelationalExpression(e, ctx):

    expr = e.expr
    other = e.other
    op = e.op

    # because of the way the add-expr production handled operator precedence
    # we sometimes have nothing to do
    if other is None:
        return expr

    ops = dict(
        [
            (">", lambda x, y: x.__gt__(y)),
            ("<", lambda x, y: x.__lt__(y)),
            ("=", lambda x, y: x.eq(y)),
            ("!=", lambda x, y: x.neq(y)),
            (">=", lambda x, y: x.__ge__(y)),
            ("<=", lambda x, y: x.__le__(y)),
            ("IN", pyop.contains),
            ("NOT IN", lambda x, y: not pyop.contains(x, y)),
        ]
    )

    if op in ("IN", "NOT IN"):

        res = op == "NOT IN"

        error = False

        if other == RDF.nil:
            other = []

        for x in other:
            try:
                if x == expr:
                    return Literal(True ^ res)
            except SPARQLError as e:
                error = e
        if not error:
            return Literal(False ^ res)
        else:
            raise error

    if op not in ("=", "!=", "IN", "NOT IN"):
        if not isinstance(expr, Literal):
            raise SPARQLError(
                "Compare other than =, != of non-literals is an error: %r" % expr
            )
        if not isinstance(other, Literal):
            raise SPARQLError(
                "Compare other than =, != of non-literals is an error: %r" % other
            )
    else:
        if not isinstance(expr, Node):
            raise SPARQLError("I cannot compare this non-node: %r" % expr)
        if not isinstance(other, Node):
            raise SPARQLError("I cannot compare this non-node: %r" % other)

    if isinstance(expr, Literal) and isinstance(other, Literal):

        if (
            expr.datatype is not None
            and expr.datatype not in XSD_DTs
            and other.datatype is not None
            and other.datatype not in XSD_DTs
        ):
            # in SPARQL for non-XSD DT Literals we can only do =,!=
            if op not in ("=", "!="):
                raise SPARQLError("Can only do =,!= comparisons of non-XSD Literals")

    try:
        r = ops[op](expr, other)
        if r == NotImplemented:
            raise SPARQLError("Error when comparing")
    except TypeError as te:
        raise SPARQLError(*te.args)
    return Literal(r)


def ConditionalAndExpression(e, ctx):

    # TODO: handle returned errors

    expr = e.expr
    other = e.other

    # because of the way the add-expr production handled operator precedence
    # we sometimes have nothing to do
    if other is None:
        return expr

    return Literal(all(EBV(x) for x in [expr] + other))


def ConditionalOrExpression(e, ctx):

    # TODO: handle errors

    expr = e.expr
    other = e.other

    # because of the way the add-expr production handled operator precedence
    # we sometimes have nothing to do
    if other is None:
        return expr
    # A logical-or that encounters an error on only one branch
    # will return TRUE if the other branch is TRUE and an error
    # if the other branch is FALSE.
    error = None
    for x in [expr] + other:
        try:
            if EBV(x):
                return Literal(True)
        except SPARQLError as e:
            error = e
    if error:
        raise error
    return Literal(False)


def not_(arg):
    return Expr("UnaryNot", UnaryNot, expr=arg)


def and_(*args):
    if len(args) == 1:
        return args[0]

    return Expr(
        "ConditionalAndExpression",
        ConditionalAndExpression,
        expr=args[0],
        other=list(args[1:]),
    )


TrueFilter = Expr("TrueFilter", lambda _1, _2: Literal(True))


def simplify(expr):
    if isinstance(expr, ParseResults) and len(expr) == 1:
        return simplify(expr[0])

    if isinstance(expr, (list, ParseResults)):
        return list(map(simplify, expr))
    if not isinstance(expr, CompValue):
        return expr
    if expr.name.endswith("Expression"):
        if expr.other is None:
            return simplify(expr.expr)

    for k in expr.keys():
        expr[k] = simplify(expr[k])
        # expr['expr']=simplify(expr.expr)
        #    expr['other']=simplify(expr.other)

    return expr


def literal(s):
    if not isinstance(s, Literal):
        raise SPARQLError("Non-literal passed as string: %r" % s)
    return s


def datetime(e):
    if not isinstance(e, Literal):
        raise SPARQLError("Non-literal passed as datetime: %r" % e)
    if not e.datatype == XSD.dateTime:
        raise SPARQLError("Literal with wrong datatype passed as datetime: %r" % e)
    return e.toPython()


def date(e) -> py_datetime.date:
    if not isinstance(e, Literal):
        raise SPARQLError("Non-literal passed as date: %r" % e)
    if e.datatype not in (XSD.date, XSD.dateTime):
        raise SPARQLError("Literal with wrong datatype passed as date: %r" % e)
    result = e.toPython()
    if isinstance(result, py_datetime.datetime):
        return result.date()
    return result


def string(s):
    """
    Make sure the passed thing is a string literal
    i.e. plain literal, xsd:string literal or lang-tagged literal
    """
    if not isinstance(s, Literal):
        raise SPARQLError("Non-literal passes as string: %r" % s)
    if s.datatype and s.datatype != XSD.string:
        raise SPARQLError("Non-string datatype-literal passes as string: %r" % s)
    return s


def numeric(expr):
    """
    return a number from a literal
    http://www.w3.org/TR/xpath20/#promotion
    or TypeError
    """

    if not isinstance(expr, Literal):
        raise SPARQLTypeError("%r is not a literal!" % expr)

    if expr.datatype not in (
        XSD.float,
        XSD.double,
        XSD.decimal,
        XSD.integer,
        XSD.nonPositiveInteger,
        XSD.negativeInteger,
        XSD.nonNegativeInteger,
        XSD.positiveInteger,
        XSD.unsignedLong,
        XSD.unsignedInt,
        XSD.unsignedShort,
        XSD.unsignedByte,
        XSD.long,
        XSD.int,
        XSD.short,
        XSD.byte,
    ):
        raise SPARQLTypeError("%r does not have a numeric datatype!" % expr)

    return expr.toPython()


def dateTimeObjects(expr):
    """
    return a dataTime/date/time/duration/dayTimeDuration/yearMonthDuration python objects from a literal
    """
    return expr.toPython()


def isCompatibleDateTimeDatatype(obj1, dt1, obj2, dt2):
    """
    Returns a boolean indicating if first object is compatible
    with operation(+/-) over second object.
    """
    if dt1 == XSD.date:
        if dt2 == XSD.yearMonthDuration:
            return True
        elif dt2 == XSD.dayTimeDuration or dt2 == XSD.Duration:
            # checking if the dayTimeDuration has no Time Component
            # else it wont be compatible with Date Literal
            if "T" in str(obj2):
                return False
            else:
                return True

    if dt1 == XSD.time:
        if dt2 == XSD.yearMonthDuration:
            return False
        elif dt2 == XSD.dayTimeDuration or dt2 == XSD.Duration:
            # checking if the dayTimeDuration has no Date Component
            # (by checking if the format is "PT...." )
            # else it wont be compatible with Time Literal
            if "T" == str(obj2)[1]:
                return True
            else:
                return False

    if dt1 == XSD.dateTime:
        # compatible with all
        return True


def calculateDuration(obj1, obj2):
    """
    returns the duration Literal between two datetime
    """
    date1 = obj1
    date2 = obj2
    difference = date1 - date2
    return Literal(difference, datatype=XSD.duration)


def calculateFinalDateTime(obj1, dt1, obj2, dt2, operation):
    """
    Calculates the final dateTime/date/time resultant after addition/
    subtraction of duration/dayTimeDuration/yearMonthDuration
    """

    # checking compatibility of datatypes (duration types and date/time/dateTime)
    if isCompatibleDateTimeDatatype(obj1, dt1, obj2, dt2):
        # proceed
        if operation == "-":
            ans = obj1 - obj2
            return Literal(ans, datatype=dt1)
        else:
            ans = obj1 + obj2
            return Literal(ans, datatype=dt1)

    else:
        raise SPARQLError("Incompatible Data types to DateTime Operations")


def EBV(rt):
    """
    Effective Boolean Value (EBV)
    * If the argument is a typed literal with a datatype of xsd:boolean,
      the EBV is the value of that argument.
    * If the argument is a plain literal or a typed literal with a
      datatype of xsd:string, the EBV is false if the operand value
      has zero length; otherwise the EBV is true.
    * If the argument is a numeric type or a typed literal with a datatype
      derived from a numeric type, the EBV is false if the operand value is
      NaN or is numerically equal to zero; otherwise the EBV is true.
    * All other arguments, including unbound arguments, produce a type error.
    """

    if isinstance(rt, Literal):

        if rt.datatype == XSD.boolean:
            return rt.toPython()

        elif rt.datatype == XSD.string or rt.datatype is None:
            return len(rt) > 0

        else:
            pyRT = rt.toPython()

            if isinstance(pyRT, Literal):
                # Type error, see: http://www.w3.org/TR/rdf-sparql-query/#ebv
                raise SPARQLTypeError(
                    "http://www.w3.org/TR/rdf-sparql-query/#ebv - ' + \
                    'Could not determine the EBV for : %r"
                    % rt
                )
            else:
                return bool(pyRT)

    else:
        raise SPARQLTypeError(
            "http://www.w3.org/TR/rdf-sparql-query/#ebv - ' + \
            'Only literals have Boolean values! %r"
            % rt
        )


def _lang_range_check(range, lang):
    """
    Implementation of the extended filtering algorithm, as defined in point
    3.3.2, of U{RFC 4647<http://www.rfc-editor.org/rfc/rfc4647.txt>}, on
    matching language ranges and language tags.
    Needed to handle the C{rdf:PlainLiteral} datatype.
    @param range: language range
    @param lang: language tag
    @rtype: boolean
        @author: U{Ivan Herman<a href="http://www.w3.org/People/Ivan/">}
        Taken from `RDFClosure/RestrictedDatatype.py`__
    .. __:http://dev.w3.org/2004/PythonLib-IH/RDFClosure/RestrictedDatatype.py
    """

    def _match(r, l_):
        """
        Matching of a range and language item: either range is a wildcard
        or the two are equal
        @param r: language range item
        @param l_: language tag item
        @rtype: boolean
        """
        return r == "*" or r == l_

    rangeList = range.strip().lower().split("-")
    langList = lang.strip().lower().split("-")
    if not _match(rangeList[0], langList[0]):
        return False
    if len(rangeList) > len(langList):
        return False

    return all(_match(*x) for x in zip(rangeList, langList))

 


#
# The MIT License (MIT)
#
# Copyright (c) 2022 by Maxime Lefrançois
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# Project      : ldpython-parser; Linked-Data Python to Python/Micropython Rewriter
#                https://gitlab.com/coswot/linked-data-python/ldpy
# Developed by : Maxime Lefrançois, maxime.lefrancois@emse.fr
#
import secrets
import rdflib
from antlr4 import TerminalNode, Token
from ldpy.rewriter.antlr.LDPythonVisitor import LDPythonVisitor
from ldpy.rewriter.antlr.LDPythonParser import LDPythonParser
from ldpy.rewriter.IndentedStringWriter import IndentedStringWriter
from ldpy.rewriter.Result import Result

SIZE = 3

# This class defines a complete generic visitor for a parse tree produced by LDPythonParser.
class LDPythonRewriter(LDPythonVisitor):
    
    def __init__(self, output:IndentedStringWriter=IndentedStringWriter()):
        self.base:str = None;
        self.namespaces = dict()
        self.output = output

    def visit(self, tree):
        return tree.accept(self)

    def visitChildren(self, node, asarray=False):
        n = node.getChildCount()
        if n==1:
            return node.getChild(0).accept(self)        
        result = Result()
        out = []
        for i in range(n):
            child = node.getChild(i)
            if isinstance(child, LDPythonParser.SuiteContext):
                raise Exception("not to be used when suite!")
            childResult = child.accept(self)
            out.append(childResult.out)
            result.stmts += childResult.stmts
            result.post_stmts += childResult.post_stmts
        if asarray:
            result.out = out
        else:
            result.out = ' '.join(out)
        return result
    
    def visitTerminal(self, node):
        return Result(node.getText())

    # Visit a parse tree produced by LDPythonParser#single_input.
    def visitSingle_input(self, ctx:LDPythonParser.Single_inputContext) -> None:
        if isinstance(ctx.getChild(0), TerminalNode):
            self.output.write("")
            return
        ctx.getChild(0).accept(self)
        if ctx.getChildCount()>1 and isinstance(ctx.getChild(0), TerminalNode):
            self.output.write("")


    # Visit a parse tree produced by LDPythonParser#file_input.
    def visitFile_input(self, ctx:LDPythonParser.File_inputContext) -> None:
        self.output.write("import rdflib")
        n = ctx.getChildCount()
        for i in range(n-1): # ignore the last token which is <EOF>
            child = ctx.getChild(i)
            if isinstance(child, TerminalNode):
                self.output.write("")
            else:
                child.accept(self)
            self.output.write("")


    # Visit a parse tree produced by LDPythonParser#eval_input.
    def visitEval_input(self, ctx:LDPythonParser.Eval_inputContext) -> None:
        raise Exception("not implemented yet")


    # Visit a parse tree produced by LDPythonParser#decorators.
    def visitDecorators(self, ctx:LDPythonParser.DecoratorsContext):
        result = Result()
        result.out = []
        for decorator in ctx.decorator():
            decoratorResult = decorator.accept(self)
            result.out.append(decoratorResult.out)
            result.stmts += decoratorResult.stmts
        return result
        

    # Visit a parse tree produced by LDPythonParser#decorated.
    def visitDecorated(self, ctx:LDPythonParser.DecoratedContext):
        result_decorators = ctx.decorators().accept(self)
        for stmt in result_decorators.stmts:
            self.output.write(stmt)
        ctx.getChild(1).decorators = result_decorators.out
        ctx.getChild(1).accept(self)
        

    # Visit a parse tree produced by LDPythonParser#base_stmt.
    def visitBase_stmt(self, ctx:LDPythonParser.Base_stmtContext) -> None:
        base = ctx.getChild(1).getText()[1:-1]
        self.base = base
        self.output.write(f"# @base <{base}> .")

    

    # Visit a parse tree produced by LDPythonParser#prefix_stmt.
    def visitPrefix_stmt(self, ctx:LDPythonParser.Prefix_stmtContext) -> None:
        prefix = ctx.NAME().getText() if ctx.NAME() else ""
        uri = ctx.IRIREF().getText()[1:-1]
        self.namespaces[prefix] = rdflib.Namespace(uri)
        self.output.write(f"# @prefix {prefix}: <{uri}> .")
    

    # Visit a parse tree produced by LDPythonParser#async_funcdef.
    def visitAsync_funcdef(self, ctx:LDPythonParser.Async_funcdefContext):
        funcdef = ctx.funcdef()
        funcdef.isasync = True
        if hasattr(ctx, "decorators"):
            funcdef.decorators = ctx.decorators
        funcdef.accept(self)


    # Visit a parse tree produced by LDPythonParser#funcdef.
    def visitFuncdef(self, ctx:LDPythonParser.FuncdefContext):
        out = "async " if hasattr(ctx, "isasync") else ""
        result_parameters = ctx.parameters().accept(self)
        out += "def " + ctx.NAME().getText() + result_parameters.out
        if ctx.test():
            result_test = ctx.test().accept(self)
            for stmt in result_test.stmts:
                self.output.write(stmt)
            out += " -> " + result_test.out
        out += ":"
        if hasattr(ctx, "decorators"):
            for decorator in ctx.decorators:
                self.output.write(decorator)
        self.output.write(out)
        # WARNING: les valeurs par défaut des paramètres doivent être à l'intérieur de la fonction, pas à l'exterieur
        # if param = None:
        #     param = param_default
        self.output.indent()
        for stmt in result_parameters.post_stmts:
            self.output.write(stmt)
        self.output.dedent()
        ctx.suite().accept(self)        


    # Visit a parse tree produced by LDPythonParser#typedargslist.
    def visitTypedargslist(self, ctx:LDPythonParser.TypedargslistContext):
        n = ctx.getChildCount()
        result = Result()
        out = []
        i=0
        # check if a test has stmts. make them post_stmts and delete '='
        while i<n:
            child = ctx.getChild(i)
            if isinstance(child, LDPythonParser.TfpdefContext):
                childResult = child.accept(self)
                out.append(childResult.out)
                name = childResult.out.split(":")[0]
                result.stmts += childResult.stmts                
            elif isinstance(child, TerminalNode) and ( child.getText() == '*' or child.getText() == '**') :
                star = child.getText()
                i+=1
                child = ctx.getChild(i)
                childResult = child.accept(self)
                out.append(star + childResult.out)
                name = childResult.out.split(":")[0]
                result.stmts += childResult.stmts
            elif isinstance(child, TerminalNode) and child.getText() == '=':
                child_testResult = ctx.getChild(i+1).accept(self)
                # if child_testResult.stmts:
                if True:
                    secret = "param_" + secrets.token_hex(SIZE)
                    result.post_stmts.append(f"if {name} == '{secret}':")
                    for stmt in child_testResult.stmts:
                        result.post_stmts.append(f"    {stmt}")
                    result.post_stmts.append(f"    {name} = {child_testResult.out}")
                    out[-1] += f"= '{secret}'"
                    i += 2
                # else:
                #     out[-1] += '=' + child_testResult.out
            i += 1
        result.out = ', '.join(out)
        return result


    # Visit a parse tree produced by LDPythonParser#simple_stmt.
    def visitSimple_stmt(self, ctx:LDPythonParser.Simple_stmtContext) -> None:
        # transforms small_stmt into simple_stmt
        out = ''
        stmts = 0
        for small_stmt in ctx.small_stmt():
            childResult = small_stmt.accept(self) # type: Result
            if childResult.stmts:
                if stmts !=0:
                    self.output.write(out)
                    out = ''
                    stmts = 0
                for stmt in childResult.stmts:
                    self.output.write(stmt)
                self.output.write(childResult.out)
            else:
                if stmts !=0:
                    out += '; '
                out += childResult.out
                stmts += 1
        if stmts !=0:
            self.output.write(out)
        # self.output.write("") # NEWLINE


    # Visit a parse tree produced by LDPythonParser#del_stmt.
    def visitDel_stmt(self, ctx:LDPythonParser.Del_stmtContext):
        childResult = ctx.exprlist().accept(self)
        result = Result('del ' + childResult.out)
        result.stmts = childResult.stmts 
        return result


    # Visit a parse tree produced by LDPythonParser#return_stmt.
    def visitReturn_stmt(self, ctx:LDPythonParser.Return_stmtContext):
        if not ctx.testlist():
            return Result("return")
        childResult = ctx.testlist().accept(self)
        result = Result('return ' + childResult.out)
        result.stmts = childResult.stmts 
        return result


    # Visit a parse tree produced by LDPythonParser#raise_stmt.
    def visitRaise_stmt(self, ctx:LDPythonParser.Raise_stmtContext):
        result = ctx.test()[0].accept(self)
        result.out = 'raise ' + result.out
        if len(ctx.test()>1):
            result2 = ctx.test()[1].accept(self)
            result.out += ' from ' + result2.out
            result.stmts += result2.stmts
        return result


    # Visit a parse tree produced by LDPythonParser#import_name.
    def visitImport_name(self, ctx:LDPythonParser.Import_nameContext):
        return Result("import " + ctx.dotted_as_names().accept(self).out)


    # Visit a parse tree produced by LDPythonParser#import_from.
    def visitImport_from(self, ctx:LDPythonParser.Import_fromContext):
        return Result(' '.join([ c.accept(self).out for c in ctx.getChildren() ]))


    # Visit a parse tree produced by LDPythonParser#import_as_name.
    def visitImport_as_name(self, ctx:LDPythonParser.Import_as_nameContext) -> Result:
        return Result(' '.join([ c.getText() for c in ctx.getChildren() ]))


    # Visit a parse tree produced by LDPythonParser#dotted_as_name.
    def visitDotted_as_name(self, ctx:LDPythonParser.Dotted_as_nameContext) -> Result:
        result = ctx.dotted_name().accept(self)
        if ctx.AS():
            result.out += ' as ' + ctx.NAME().getText()
        return result

    # Visit a parse tree produced by LDPythonParser#import_as_names.
    def visitImport_as_names(self, ctx:LDPythonParser.Import_as_namesContext) -> Result:
        return Result(', '.join([ c.accept(self).out for c in ctx.import_as_name() ]))


    # Visit a parse tree produced by LDPythonParser#dotted_as_names.
    def visitDotted_as_names(self, ctx:LDPythonParser.Dotted_as_namesContext) -> Result:
        return Result(', '.join([ c.accept(self).out for c in ctx.dotted_as_name() ]))


    # Visit a parse tree produced by LDPythonParser#dotted_name.
    def visitDotted_name(self, ctx:LDPythonParser.Dotted_nameContext) -> Result:
        return Result('.'.join([ c.getText() for c in ctx.NAME() ]))
    

    # Visit a parse tree produced by LDPythonParser#global_stmt.
    def visitGlobal_stmt(self, ctx:LDPythonParser.Global_stmtContext) -> Result:
        return Result('global ' + ', '.join([ c.getText() for c in ctx.NAME() ]))


    # Visit a parse tree produced by LDPythonParser#nonlocal_stmt.
    def visitNonlocal_stmt(self, ctx:LDPythonParser.Nonlocal_stmtContext) -> Result:
        return Result('nonlocal ' + ', '.join([ c.getText() for c in ctx.NAME() ]))


    # Visit a parse tree produced by LDPythonParser#async_stmt.
    def visitAsync_stmt(self, ctx:LDPythonParser.Async_stmtContext):
        child = ctx.getChild(1)
        child.isasync = True
        child.accept(self)


    # Visit a parse tree produced by LDPythonParser#if_stmt.
    def visitIf_stmt(self, ctx:LDPythonParser.If_stmtContext):
        stmts = []
        # the if test
        result_if = ctx.test()[0].accept(self)
        stmts += result_if.stmts
        # the elifs tests
        results_elif = [c.accept(self) for c in ctx.test()[1:]]
        for result_elif in results_elif:
            stmts += result_elif.stmts
        # print the stmts
        for stmt in stmts:
            self.output.write(stmt)
        # the if
        self.output.write("if " + result_if.out + ":")
        ctx.suite()[0].accept(self)
        # the elifs
        for i in range(len(results_elif)):
            self.output.write("elif " + results_elif[i].out + ":")
            ctx.suite()[i+1].accept(self)
        # the else
        if ctx.ELSE():
            self.output.write("else:")
            ctx.suite()[-1].accept(self)


    # Visit a parse tree produced by LDPythonParser#while_stmt.
    def visitWhile_stmt(self, ctx:LDPythonParser.While_stmtContext):
        result_test = ctx.test().accept(self)
        for stmt in result_test.stmts:
            self.output.write(stmt)
        self.output.write("while " + result_test.out + ":")
        ctx.suite()[0].accept(self)        
        if len(ctx.suite())>1:
            self.output.write("else:")
            ctx.suite()[1].accept(self)


    # Visit a parse tree produced by LDPythonParser#for_stmt.
    def visitFor_stmt(self, ctx:LDPythonParser.For_stmtContext):
        out = "async " if hasattr(ctx, "isasync") else ""
        exprlist = ctx.exprlist().accept(self)
        for stmt in exprlist.stmts:
            self.output.write(stmt)
        testlist = ctx.testlist().accept(self)
        for stmt in testlist.stmts:
            self.output.write(stmt)
        out += "for " + exprlist.out + " in " + testlist.out + ":"
        self.output.write(out)
        ctx.suite()[0].accept(self)        
        if len(ctx.suite())>1:
            self.output.write("else:")
            ctx.suite()[1].accept(self)


    # Visit a parse tree produced by LDPythonParser#try_stmt.
    def visitTry_stmt(self, ctx:LDPythonParser.Try_stmtContext):
        results_except_clause = [c.accept(self) for c in ctx.except_clause()]
        for result_except_clause in results_except_clause:
            for stmt in result_except_clause.stmts:
                self.output.write(stmt)
        self.output.write("try:")
        ctx.suite()[0].accept(self)
        n = len(results_except_clause)
        for i in range(n):
            self.output.write(results_except_clause[i].out + ":")
            ctx.suite()[i+1].accept(self)
        for i in range(ctx.getChildCount()-7,ctx.getChildCount()):
            if ctx.getChild(i) == ctx.ELSE():
                self.output.write("else:")
                ctx.getChild(i+2).accept(self)
            elif ctx.getChild(i) == ctx.FINALLY():
                self.output.write("finally:")
                ctx.getChild(i+2).accept(self)
                

    # Visit a parse tree produced by LDPythonParser#with_stmt.
    def visitWith_stmt(self, ctx:LDPythonParser.With_stmtContext):
        out = "async " if hasattr(ctx, "isasync") else ""
        results_with_item = [c.accept(self) for c in ctx.with_item()]
        results_with_item_out = [r.out for r in results_with_item]
        for result_with_item in results_with_item:
            for stmt in result_with_item.stmts:
                self.output.write(stmt)
        out += "with " + ",".join(results_with_item_out)
        self.output.write(out)
        ctx.suite().accept(self)


    # Visit a parse tree produced by LDPythonParser#suite.
    def visitSuite(self, ctx:LDPythonParser.SuiteContext):
        self.output.indent()
        if ctx.simple_stmt():
            ctx.simple_stmt.accept(self)
        for stmt in ctx.stmt():
            stmt.accept(self)
        self.output.dedent()


    # Visit a parse tree produced by LDPythonParser#classdef.
    def visitClassdef(self, ctx:LDPythonParser.ClassdefContext):
        out = []
        n = ctx.getChildCount()
        for i in range(n-1):
            childResult = ctx.getChild(i).accept(self)
            out.append(childResult.out)
            for stmt in childResult.stmts:
                self.output.write(stmt)
        if hasattr(ctx, "decorators"):
            for decorator in ctx.decorators:
                self.output.write(decorator)
        self.output.write(' '.join(out))
        ctx.suite().accept(self)


    # Visit a parse tree produced by LDPythonParser#construct_template.
    def visitConstruct_template(self, ctx:LDPythonParser.Construct_templateContext):
        rdfgraph = "graph_" + secrets.token_hex(SIZE)
        result = Result(rdfgraph)

        base = f"base='{self.base}'" if self.base else ""
        result.stmts.append(f"{rdfgraph} = rdflib.Graph({base})")

        for prefix, uri in self.namespaces.items():
            result.stmts.append(f"{rdfgraph}.namespace_manager.bind('{prefix}', '{uri}')")

        if ctx.construct_triples():
            childResult = ctx.construct_triples().accept(self)
            for stmt in childResult.stmts:
                if isinstance(stmt, str):
                    result.stmts.append(stmt)
                elif isinstance(stmt, tuple):
                    result.stmts.append(f"{rdfgraph}.add(({stmt[0]},{stmt[1]},{stmt[2]}))")

        return result

    # Visit a parse tree produced by LDPythonParser#construct_triples.
    def visitConstruct_triples(self, ctx:LDPythonParser.Construct_triplesContext):
        result = Result()
        
        childResult = ctx.triples_same_subject().accept(self)
        result.add_stmts(childResult)
        
        if ctx.construct_triples():
            childResult = ctx.construct_triples().accept(self)
            result.add_stmts(childResult)

        return result


    # Visit a parse tree produced by LDPythonParser#triples_same_subject.
    def visitTriples_same_subject(self, ctx:LDPythonParser.Triples_same_subjectContext):
        result = Result()
        
        subjectResult = ctx.getChild(0).accept(self)
        subject = subjectResult.out
        result.add_stmts(subjectResult)

        predicateObjectResult = ctx.getChild(1).accept(self)
        result.add_stmts(predicateObjectResult, subject)
        
        return result


    # Visit a parse tree produced by LDPythonParser#property_list.
    def visitProperty_list(self, ctx:LDPythonParser.Property_listContext):
        if ctx.property_list_not_empty():
            return ctx.property_list_not_empty().accept(self)
        else:
            return Result()


    # Visit a parse tree produced by LDPythonParser#property_list_not_empty.
    def visitProperty_list_not_empty(self, ctx:LDPythonParser.Property_list_not_emptyContext):
        result = Result()
        i = 0
        n = ctx.getChildCount()
        while i < n:
            if isinstance(ctx.getChild(i), LDPythonParser.VerbContext):
                verbResult = ctx.getChild(i).accept(self)
                verb = verbResult.out
                result.add_stmts(verbResult)
                i += 1
                object_listResult = ctx.getChild(i).accept(self)
                result.add_stmts(object_listResult, verb)
            i += 1                
        return result


    # Visit a parse tree produced by LDPythonParser#verb.
    def visitVerb(self, ctx:LDPythonParser.VerbContext):
        if isinstance(ctx.getChild(0), TerminalNode):
            return Result("rdflib.RDF.type")
        else:
            return ctx.getChild(0).accept(self)


    # Visit a parse tree produced by LDPythonParser#object_list.
    def visitObject_list(self, ctx:LDPythonParser.Object_listContext):
        result = Result()
        for child in ctx.object_():
            childResult = child.accept(self)
            result.add_stmts(childResult)
            result.stmts.append((childResult.out,))
        return result


    # Visit a parse tree produced by LDPythonParser#blank_node_property_list.
    def visitBlank_node_property_list(self, ctx:LDPythonParser.Blank_node_property_listContext):
        rdfsubject = "bn_" + secrets.token_hex(SIZE)
        result = Result(rdfsubject)
        result.stmts.append(f"{rdfsubject} = rdflib.BNode()")
        childResult = ctx.property_list_not_empty().accept(self)
        for stmt in childResult.stmts:
            if isinstance(stmt, tuple) and len(stmt)<3:
                result.stmts.append((rdfsubject, *stmt))
            else:
                result.stmts.append(stmt)
        return result
    

    # Visit a parse tree produced by LDPythonParser#collection.
    def visitCollection(self, ctx:LDPythonParser.CollectionContext):
        secret = secrets.token_hex(SIZE)
        i = 0
        rdflist = f"list_{secret}_{i}"
        result = Result(rdflist)
        result.stmts.append(f"{rdflist} = rdflib.BNode()")
        n = len(ctx.graph_node())
        while i < n:
            rdflist = f"list_{secret}_{i}"
            childResult = ctx.graph_node()[i].accept(self)
            result.add_stmts(childResult)
            result.stmts.append((rdflist, "rdflib.RDF.first", childResult.out))
            if i < n-1:
                rdfrest = f"list_{secret}_{i+1}"
                result.stmts.append(f"{rdfrest} = rdflib.BNode()")
                result.stmts.append((rdflist, "rdflib.RDF.rest", rdfrest))
            else:
                result.stmts.append((rdflist, "rdflib.RDF.rest", "rdflib.RDF.nil"))
            i += 1
        return result


    # Visit a parse tree produced by LDPythonParser#var.
    def visitVar(self, ctx:LDPythonParser.VarContext):
        if ctx.VAR1() or ctx.VAR2():
            varName = ctx.getText()[1:]
            return Result(f"rdflib.Variable('{varName}')")
        return ctx.getChild(0).accept(self)


    # Visit a parse tree produced by LDPythonParser#graph_term.
    def visitGraph_term(self, ctx:LDPythonParser.Graph_termContext):
        if ctx.NUMBER() or ctx.TRUE() or ctx.FALSE():
            return Result(f"rdflib.Literal({ctx.getChild(0).getText()})")
        return ctx.getChild(0).accept(self)


    # Visit a parse tree produced by LDPythonParser#rdf_literal.
    def visitRdf_literal(self, ctx:LDPythonParser.Rdf_literalContext):
        string = ctx.STRING().getText()
        # if not ctx.iri() and not ctx.LANGTAG():
        #     return Result(string)
        if ctx.iri():
            datatype = ctx.iri().accept(self)
            return Result(f"rdflib.Literal({string},datatype={datatype})")
        if ctx.firi() or ctx.fnode():
            datatype = ctx.getChild(2).accept(self)
            result = Result(f"rdflib.Literal({string},datatype={datatype})")
            result.add_stmts(datatype)
            return result
        elif ctx.LANGTAG():
            lang = ctx.LANGTAG().getText()[1:]
            return Result(f"rdflib.Literal({string},lang='{lang}')")
        else:
            return Result(f"rdflib.Literal({string})")
        

    # Visit a parse tree produced by LDPythonParser#iri.
    def visitIri(self, ctx:LDPythonParser.IriContext):
        if ctx.IRIREF():
            uri = ctx.getText()[1:-1]
            if self.base:
                return Result(f"rdflib.URIRef('{uri}', '{self.base}')")
            else:
                return Result(f"rdflib.URIRef('{uri}')")
        else:
            return ctx.prefixed_name().accept(self)

    # Visit a parse tree produced by LDPythonParser#prefixed_name.
    def visitPrefixed_name(self, ctx:LDPythonParser.Prefixed_nameContext):
        prefix , localName  = ctx.getText().split(":", 1)
        if prefix not in set(self.namespaces):
            offendingToken : Token = ctx.start 
            line = offendingToken.line
            column = offendingToken.column
            raise Exception(f"Error at {line}:{column}: prefix {prefix} is undefined")
        uri = self.namespaces[prefix][localName]
        if self.base:
            return Result(f"rdflib.URIRef('{uri}', '{self.base}')")
        else:
            return Result(f"rdflib.URIRef('{uri}')")


    # Visit a parse tree produced by LDPythonParser#blank_node.
    def visitBlank_node(self, ctx:LDPythonParser.Blank_nodeContext):
        name = "bn_" + secrets.token_hex(SIZE)
        result = Result(name)
        if ctx.BLANK_NODE_LABEL():
            label = ctx.getText()[2:]
            result.stmts.append(f"{name} = rdflib.BNode('{label}')")
        else:
            result.stmts.append(f"{name} = rdflib.BNode()")
        
        return result    


    # Visit a parse tree produced by LDPythonParser#nil.
    def visitNil(self, ctx:LDPythonParser.NilContext):
        return Result("rdflib.RDF.nil")


    # Visit a parse tree produced by LDPythonParser#anon.
    def visitAnon(self, ctx:LDPythonParser.AnonContext):
        name = "bn_" + secrets.token_hex(SIZE)
        result = Result(name)
        result.stmts.append(f"{name} = rdflib.BNode()")
        return result
    
    
    # Visit a parse tree produced by LDPythonParser#firi.
    def visitFiri(self, ctx:LDPythonParser.FiriContext):
        name = "var_" + secrets.token_hex(SIZE)
        result = Result(name)
        out = []
        string = ctx.FIRIREF_START().getText()[2:-1]
        if len(string)>0:
            out.append(f"'{string}'")
        n = ctx.getChildCount()
        for i in range(1,n,2):
            testResult = ctx.getChild(i).accept(self)
            result.add_stmts(testResult)
            out.append("str(" + testResult.out + ")")
            string = ctx.getChild(i+1).accept(self).out[1:-1]
            if len(string)>0:
                out.append(f"'{string}'")
        iri = ' + '.join(out)
        result.stmts.append(f"{name} = rdflib.URIRef({iri})")
        return result

    # Visit a parse tree produced by LDPythonParser#fnode.
    def visitFnode(self, ctx:LDPythonParser.FnodeContext):
        testResult = ctx.test().accept(self)
        if not testResult.stmts:
            return testResult
        name = "var_" + secrets.token_hex(SIZE)
        result = Result(name)
        result.add_stmts(testResult)
        result.stmts.append(f"{name} = {testResult.out}")
        return result



# from https://github.com/charmoniumQ/antlr4-python-grun
from typing import TypeVar, Iterator, Tuple, Iterable

Value = TypeVar("Value")

def first_sentinel(iterable: Iterable[Value]) -> Iterator[Tuple[Value, bool]]:
    iterator = iter(iterable)
    yield (next(iterator), True)
    for elem in iterator:
        yield (elem, False)




from typing import Union, Dict, List, Set
from rdflib import RDF, OWL, Graph, URIRef, BNode, Literal, Variable
from rdflib.term import Node

RDFTerm = Union[URIRef, BNode, Literal]
Term = Union[Variable, URIRef, BNode, Literal]
RDFInstanceMapping = Dict[BNode, RDFTerm]
SolutionMapping = Dict[Variable, RDFTerm]

def instantiateBGP(input:Graph, solutionMappings, initialGraph:Graph=None):
    if initialGraph == None:
        initialGraph=Graph(base=input.base)
    initialGraph.namespace_manager = input.namespace_manager
    if solutionMappings is None:
        return initialGraph
    if isinstance(solutionMappings, dict):
        solutionMappings = [solutionMappings]
    if not isinstance(solutionMappings, list):
        raise AssertionError("solutionMappings needs to be a dictionary, or list of dictionaries")
    def instantiateTerm(t:Term, solutionMapping:SolutionMapping, rdfInstanceMapping:RDFInstanceMapping):
        """If the input term is a variable, replace it by its value in the solution mapping. Return None if this variable is not mapped. If the input term is a blank node, replace it by its value in the RDF instance mapping. Create one if it is not mapped."""
        if isinstance(t, Variable):
            if t in set(solutionMapping):
                value = solutionMapping[t]
                if value == None:
                    return None
                if not isinstance(value, Node):
                    value = Literal(value)
                return value
            else:
                return None
        if isinstance(t, BNode):
            if t in set(rdfInstanceMapping):
                return rdfInstanceMapping[t]
            else:
                bnode = BNode()
                rdfInstanceMapping[t] = bnode
                return bnode
        return t            
    for solutionMapping in solutionMappings:
        rdfInstanceMapping={} # type: RDFInstanceMapping
        for s,p,o in input:
            s = instantiateTerm(s, solutionMapping, rdfInstanceMapping)
            p = instantiateTerm(p, solutionMapping, rdfInstanceMapping)
            o = instantiateTerm(o, solutionMapping, rdfInstanceMapping)
            if s and p and o:
                initialGraph.add((s, p, o))
    return initialGraph



#
# The MIT License (MIT)
#
# Copyright (c) 2022 by Maxime Lefrançois
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# Project      : ldpython-parser; Linked-Data Python to Python/Micropython Rewriter
#                https://gitlab.com/coswot/linked-data-python/ldpy
# Developed by : Maxime Lefrançois, maxime.lefrancois@emse.fr
#
import sys
from antlr4 import FileStream, Token
from antlr4.error.ErrorListener import ErrorListener
from antlr4.error.DiagnosticErrorListener import DiagnosticErrorListener
from ldpy.rewriter import *
import argparse

def debugTokens(input_stream):
    lexer = LDPythonLexer(input_stream)
    import ldpy.grun.lib as lib
    tokenTypes = lib.readTokenTypes("ldpy/rewriter/antlr/LDPython.tokens")
    printTokens(lexer, tokenTypes)

class MainErrorListener(ErrorListener):
    
    def __init__(self):
        self.errors = []
    
    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        self.errors.append("line " + str(line) + ":" + str(column) + " " + msg)


def parseTree(input_stream, diagnose=False): 
    lexer = LDPythonLexer(input_stream)
    stream = MultiChannelTokenStream(lexer)
    parser = LDPythonParser(stream)
    parser.removeErrorListeners()
    errorListener = MainErrorListener()
    parser.addErrorListener(errorListener)
    if diagnose:
        diagnosticErrorListener = DiagnosticErrorListener()
        parser.addErrorListener(diagnosticErrorListener)
    tree = parser.file_input()
    if parser.getNumberOfSyntaxErrors() != 0:
        raise SyntaxError("Exception while parsing the input:\n  " + "\n  ".join(errorListener.errors))
    return tree

def printTree(tree):
    import ldpy.grun.lib as lib
    tokenTypes = lib.readTokenTypes("ldpy/rewriter/antlr/LDPython.tokens")
    try:
        print(lib.format_tree(tree, True, tokenTypes))
    except Exception as e:
        import traceback
        print(traceback.format_exc())
    
def execResult(code): # from https://stackoverflow.com/a/47337130
    from io import StringIO
    old_stdout = sys.stdout
    redirected_output = sys.stdout = StringIO()
    try:
        exec(code, locals(), locals())
    except:
        raise 
    finally: # !
        sys.stdout = old_stdout # !

    print(redirected_output.getvalue())
    
    
def printTokens(lexer, tokenTypes):
    i = 0
    while True:
        t:Token = lexer.nextToken()
        print( f"[@{i},{t.start}:{t.stop}='{t.text}',<{t.type}:{tokenTypes[t.type]}>,{t.line}:{t.column} in channel {t.channel}]")
        i += 1
        if t.type == -1:
            break

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Rewrite a .ldpy file into an equivalent python code.')
    parser.add_argument('-l', '--debug-lexer', dest='lexer', action='store_true', default=False,
                        help='print the lexer output')
    parser.add_argument('-p', '--debug-parser', dest='parser', action='store_true', default=False,
                        help='print the parser output')
    parser.add_argument('-d', '--diagnose-syntax', dest='diagnose', action='store_true', default=False,
                        help='diagnose the ambiguities in the syntax')
    parser.add_argument('-s', '--silent', dest='silent', action='store_true', default=False,
                        help='do not display the output code')
    parser.add_argument('-w', '--write-output', dest='write', action='store_true', default=False,
                        help='write the output python code with the .py extension')
    parser.add_argument('-x', '--execute', dest='exec', action='store_true', default=False,
                        help='execute the output code')
    parser.add_argument('file', metavar='file', type=str,
                        help='the file to rewrite')
    args = parser.parse_args(sys.argv[1:])
    
    if not args.file.endswith(".ldpy"):
        raise AssertionError("the file must have the .ldpy extension")
    input_stream = FileStream(args.file)
    
    if args.lexer:
        print("Tokens are:")
        debugTokens(input_stream)
        input_stream = FileStream(args.file)

    tree = parseTree(input_stream, args.diagnose)

    
    if args.parser:
        print("Parsed tree is:")
        printTree(tree)

    output = IndentedStringWriter()
    visitor = LDPythonRewriter(output)
    visitor.visit(tree)

    if not args.silent:
        print("output:")
        print(output.getvalue())
        
    if args.write:
        with open(args.file[:-4]+"py", "w") as f:
            f.write(output.getvalue())

    if args.exec:
        print("output:")
        execResult(output.getvalue())
    
    


from ldpy.rewriter.IndentedStringWriter import IndentedStringWriter
from ldpy.rewriter.MultiChannelTokenStream import MultiChannelTokenStream
from ldpy.rewriter.antlr.LDPythonLexer import LDPythonLexer
from ldpy.rewriter.antlr.LDPythonParser import LDPythonParser
from ldpy.rewriter.antlr.LDPythonVisitor import LDPythonVisitor
from ldpy.rewriter.LDPythonRewriter import LDPythonRewriter
from ldpy.rewriter.LDPythonRewriter import LDPythonRewriter
from ldpy.ldpy import instantiateBGP

__version__ = "0.0.2"
__date__ = "2022-04-30"



#
# The MIT License (MIT)
#
# Copyright (c) 2022 by Maxime Lefrançois
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# Project      : ldpython-parser; Linked-Data Python to Python/Micropython Rewriter
#                https://gitlab.com/coswot/linked-data-python/ldpy
# Developed by : Maxime Lefrançois, maxime.lefrancois@emse.fr
#
from io import TextIOBase, StringIO

class IndentedStringWriter(TextIOBase):
    
    def __init__(self, output:TextIOBase = StringIO()):
        self._indent:int = 0
        self.output = output

    def indent(self):
        self._indent+=1
    
    def dedent(self):
        self._indent-=1
    
    def write(self, s:str="", nl=True):
        if nl==True:
            self.output.write("\n" + " " * 4 * self._indent)
        self.output.write(str(s))
        
    def getvalue(self):
        return self.output.getvalue()




from ldpy.rewriter.IndentedStringWriter import IndentedStringWriter
from ldpy.rewriter.MultiChannelTokenStream import MultiChannelTokenStream
from ldpy.rewriter.antlr.LDPythonLexer import LDPythonLexer
from ldpy.rewriter.antlr.LDPythonParser import LDPythonParser
from ldpy.rewriter.antlr.LDPythonVisitor import LDPythonVisitor
from ldpy.rewriter.LDPythonRewriter import LDPythonRewriter
from ldpy.rewriter.LDPythonRewriter import LDPythonRewriter



# from https://github.com/charmoniumQ/antlr4-python-grun

import enum
import json
import importlib
import subprocess
import shutil
from pathlib import Path
import os
from typing import Tuple, Iterable
import antlr4  # type: ignore
from . import util

def readTokenTypes(path):
    tokenTypes = dict()
    tokenTypes[-1] = "EOF"
    with open(path, 'r') as tokensFile:
        for line in tokensFile:
            i = line.rindex("=")
            tokenTypes[int(line[i+1:])] = line[:i]
    return tokenTypes

def format_tree(
    root: antlr4.ParserRuleContext,
    pretty: bool,
    tokenTypes: dict()
) -> str:
    stack = [(root, True)]
    depth = 0
    buf = []
    while stack:
        node, last = stack.pop()
        if node == "end":
            depth -= 1
            # buf.append(depth * " " if pretty else "")
            # buf.append(")")
            # if not last:
            #     buf.append(" ")
        elif isinstance(node, antlr4.tree.Tree.TerminalNodeImpl):
            buf.append(depth * " " if pretty else "")
            token = node.symbol
            lexer = token.getTokenSource()
            token_type_map = {
                symbolic_id + len(lexer.literalNames) - 1: symbolic_name
                for symbolic_id, symbolic_name in enumerate(lexer.symbolicNames)
            }
            type_ = token_type_map.get(token.type, "literal")
            buf.append(f"({tokenTypes[token.type] if token.type in tokenTypes else token.type} {token.text})")
            if not last:
                buf.append(" ")
            if pretty:
                buf.append("\n")
        else:
            buf.append(depth * " " if pretty else "")
            rule_name = node.parser.ruleNames[node.getRuleIndex()]
            buf.append(f"({rule_name}")
            depth += 1
            children = [] if node.children is None else node.children
            stack.append(("end", last))
            stack.extend(util.first_sentinel(list(children)[::-1]))
            if pretty:
                buf.append("\n")
    return "".join(buf)


# Generated from grammars/LDPython.g4 by ANTLR 4.10.1
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .LDPythonParser import LDPythonParser
else:
    from LDPythonParser import LDPythonParser

from ldpy.rewriter import MultiChannelTokenStream
from antlr4.error import Errors


# This class defines a complete generic visitor for a parse tree produced by LDPythonParser.

class LDPythonVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by LDPythonParser#single_input.
    def visitSingle_input(self, ctx:LDPythonParser.Single_inputContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#file_input.
    def visitFile_input(self, ctx:LDPythonParser.File_inputContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#eval_input.
    def visitEval_input(self, ctx:LDPythonParser.Eval_inputContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#decorator.
    def visitDecorator(self, ctx:LDPythonParser.DecoratorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#decorators.
    def visitDecorators(self, ctx:LDPythonParser.DecoratorsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#decorated.
    def visitDecorated(self, ctx:LDPythonParser.DecoratedContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#base_stmt.
    def visitBase_stmt(self, ctx:LDPythonParser.Base_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#prefix_stmt.
    def visitPrefix_stmt(self, ctx:LDPythonParser.Prefix_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#async_funcdef.
    def visitAsync_funcdef(self, ctx:LDPythonParser.Async_funcdefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#funcdef.
    def visitFuncdef(self, ctx:LDPythonParser.FuncdefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#parameters.
    def visitParameters(self, ctx:LDPythonParser.ParametersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#typedargslist.
    def visitTypedargslist(self, ctx:LDPythonParser.TypedargslistContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#tfpdef.
    def visitTfpdef(self, ctx:LDPythonParser.TfpdefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#varargslist.
    def visitVarargslist(self, ctx:LDPythonParser.VarargslistContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#vfpdef.
    def visitVfpdef(self, ctx:LDPythonParser.VfpdefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#stmt.
    def visitStmt(self, ctx:LDPythonParser.StmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#simple_stmt.
    def visitSimple_stmt(self, ctx:LDPythonParser.Simple_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#small_stmt.
    def visitSmall_stmt(self, ctx:LDPythonParser.Small_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#expr_stmt.
    def visitExpr_stmt(self, ctx:LDPythonParser.Expr_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#testlist_star_expr.
    def visitTestlist_star_expr(self, ctx:LDPythonParser.Testlist_star_exprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#augassign.
    def visitAugassign(self, ctx:LDPythonParser.AugassignContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#del_stmt.
    def visitDel_stmt(self, ctx:LDPythonParser.Del_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#pass_stmt.
    def visitPass_stmt(self, ctx:LDPythonParser.Pass_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#flow_stmt.
    def visitFlow_stmt(self, ctx:LDPythonParser.Flow_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#break_stmt.
    def visitBreak_stmt(self, ctx:LDPythonParser.Break_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#continue_stmt.
    def visitContinue_stmt(self, ctx:LDPythonParser.Continue_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#return_stmt.
    def visitReturn_stmt(self, ctx:LDPythonParser.Return_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#yield_stmt.
    def visitYield_stmt(self, ctx:LDPythonParser.Yield_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#raise_stmt.
    def visitRaise_stmt(self, ctx:LDPythonParser.Raise_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#import_stmt.
    def visitImport_stmt(self, ctx:LDPythonParser.Import_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#import_name.
    def visitImport_name(self, ctx:LDPythonParser.Import_nameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#import_from.
    def visitImport_from(self, ctx:LDPythonParser.Import_fromContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#import_as_name.
    def visitImport_as_name(self, ctx:LDPythonParser.Import_as_nameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#dotted_as_name.
    def visitDotted_as_name(self, ctx:LDPythonParser.Dotted_as_nameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#import_as_names.
    def visitImport_as_names(self, ctx:LDPythonParser.Import_as_namesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#dotted_as_names.
    def visitDotted_as_names(self, ctx:LDPythonParser.Dotted_as_namesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#dotted_name.
    def visitDotted_name(self, ctx:LDPythonParser.Dotted_nameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#global_stmt.
    def visitGlobal_stmt(self, ctx:LDPythonParser.Global_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#nonlocal_stmt.
    def visitNonlocal_stmt(self, ctx:LDPythonParser.Nonlocal_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#assert_stmt.
    def visitAssert_stmt(self, ctx:LDPythonParser.Assert_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#compound_stmt.
    def visitCompound_stmt(self, ctx:LDPythonParser.Compound_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#async_stmt.
    def visitAsync_stmt(self, ctx:LDPythonParser.Async_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#if_stmt.
    def visitIf_stmt(self, ctx:LDPythonParser.If_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#while_stmt.
    def visitWhile_stmt(self, ctx:LDPythonParser.While_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#for_stmt.
    def visitFor_stmt(self, ctx:LDPythonParser.For_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#try_stmt.
    def visitTry_stmt(self, ctx:LDPythonParser.Try_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#with_stmt.
    def visitWith_stmt(self, ctx:LDPythonParser.With_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#with_item.
    def visitWith_item(self, ctx:LDPythonParser.With_itemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#except_clause.
    def visitExcept_clause(self, ctx:LDPythonParser.Except_clauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#suite.
    def visitSuite(self, ctx:LDPythonParser.SuiteContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#test.
    def visitTest(self, ctx:LDPythonParser.TestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#test_nocond.
    def visitTest_nocond(self, ctx:LDPythonParser.Test_nocondContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#lambdef.
    def visitLambdef(self, ctx:LDPythonParser.LambdefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#lambdef_nocond.
    def visitLambdef_nocond(self, ctx:LDPythonParser.Lambdef_nocondContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#or_test.
    def visitOr_test(self, ctx:LDPythonParser.Or_testContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#and_test.
    def visitAnd_test(self, ctx:LDPythonParser.And_testContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#not_test.
    def visitNot_test(self, ctx:LDPythonParser.Not_testContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#comparison.
    def visitComparison(self, ctx:LDPythonParser.ComparisonContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#comp_op.
    def visitComp_op(self, ctx:LDPythonParser.Comp_opContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#star_expr.
    def visitStar_expr(self, ctx:LDPythonParser.Star_exprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#expr.
    def visitExpr(self, ctx:LDPythonParser.ExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#xor_expr.
    def visitXor_expr(self, ctx:LDPythonParser.Xor_exprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#and_expr.
    def visitAnd_expr(self, ctx:LDPythonParser.And_exprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#shift_expr.
    def visitShift_expr(self, ctx:LDPythonParser.Shift_exprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#arith_expr.
    def visitArith_expr(self, ctx:LDPythonParser.Arith_exprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#term.
    def visitTerm(self, ctx:LDPythonParser.TermContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#factor.
    def visitFactor(self, ctx:LDPythonParser.FactorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#power.
    def visitPower(self, ctx:LDPythonParser.PowerContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#atom_expr.
    def visitAtom_expr(self, ctx:LDPythonParser.Atom_exprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#atom.
    def visitAtom(self, ctx:LDPythonParser.AtomContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#testlist_comp.
    def visitTestlist_comp(self, ctx:LDPythonParser.Testlist_compContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#trailer.
    def visitTrailer(self, ctx:LDPythonParser.TrailerContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#subscriptlist.
    def visitSubscriptlist(self, ctx:LDPythonParser.SubscriptlistContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#subscript.
    def visitSubscript(self, ctx:LDPythonParser.SubscriptContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#sliceop.
    def visitSliceop(self, ctx:LDPythonParser.SliceopContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#exprlist.
    def visitExprlist(self, ctx:LDPythonParser.ExprlistContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#testlist.
    def visitTestlist(self, ctx:LDPythonParser.TestlistContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#dictorsetmaker.
    def visitDictorsetmaker(self, ctx:LDPythonParser.DictorsetmakerContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#classdef.
    def visitClassdef(self, ctx:LDPythonParser.ClassdefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#arglist.
    def visitArglist(self, ctx:LDPythonParser.ArglistContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#argument.
    def visitArgument(self, ctx:LDPythonParser.ArgumentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#comp_iter.
    def visitComp_iter(self, ctx:LDPythonParser.Comp_iterContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#comp_for.
    def visitComp_for(self, ctx:LDPythonParser.Comp_forContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#comp_if.
    def visitComp_if(self, ctx:LDPythonParser.Comp_ifContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#encoding_decl.
    def visitEncoding_decl(self, ctx:LDPythonParser.Encoding_declContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#yield_expr.
    def visitYield_expr(self, ctx:LDPythonParser.Yield_exprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#yield_arg.
    def visitYield_arg(self, ctx:LDPythonParser.Yield_argContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#construct_template.
    def visitConstruct_template(self, ctx:LDPythonParser.Construct_templateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#construct_triples.
    def visitConstruct_triples(self, ctx:LDPythonParser.Construct_triplesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#triples_same_subject.
    def visitTriples_same_subject(self, ctx:LDPythonParser.Triples_same_subjectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#property_list.
    def visitProperty_list(self, ctx:LDPythonParser.Property_listContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#property_list_not_empty.
    def visitProperty_list_not_empty(self, ctx:LDPythonParser.Property_list_not_emptyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#verb.
    def visitVerb(self, ctx:LDPythonParser.VerbContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#object_list.
    def visitObject_list(self, ctx:LDPythonParser.Object_listContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#object_.
    def visitObject_(self, ctx:LDPythonParser.Object_Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#triples_node.
    def visitTriples_node(self, ctx:LDPythonParser.Triples_nodeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#blank_node_property_list.
    def visitBlank_node_property_list(self, ctx:LDPythonParser.Blank_node_property_listContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#collection.
    def visitCollection(self, ctx:LDPythonParser.CollectionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#graph_node.
    def visitGraph_node(self, ctx:LDPythonParser.Graph_nodeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#var_or_term.
    def visitVar_or_term(self, ctx:LDPythonParser.Var_or_termContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#var_or_iri.
    def visitVar_or_iri(self, ctx:LDPythonParser.Var_or_iriContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#var.
    def visitVar(self, ctx:LDPythonParser.VarContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#graph_term.
    def visitGraph_term(self, ctx:LDPythonParser.Graph_termContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#rdf_literal.
    def visitRdf_literal(self, ctx:LDPythonParser.Rdf_literalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#iri.
    def visitIri(self, ctx:LDPythonParser.IriContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#prefixed_name.
    def visitPrefixed_name(self, ctx:LDPythonParser.Prefixed_nameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#all_names.
    def visitAll_names(self, ctx:LDPythonParser.All_namesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#blank_node.
    def visitBlank_node(self, ctx:LDPythonParser.Blank_nodeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#nil.
    def visitNil(self, ctx:LDPythonParser.NilContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#anon.
    def visitAnon(self, ctx:LDPythonParser.AnonContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#firi.
    def visitFiri(self, ctx:LDPythonParser.FiriContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by LDPythonParser#fnode.
    def visitFnode(self, ctx:LDPythonParser.FnodeContext):
        return self.visitChildren(ctx)



del LDPythonParser


#
# The MIT License (MIT)
#
# Copyright (c) 2022 by Maxime Lefrançois
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# Project      : ldpython-parser; Linked-Data Python to Python/Micropython Rewriter
#                https://gitlab.com/coswot/linked-data-python/ldpy
# Developed by : Maxime Lefrançois, maxime.lefrancois@emse.fr
#
from typing import List
from antlr4.Lexer import Lexer
from antlr4.Token import Token
from antlr4.BufferedTokenStream import BufferedTokenStream

# Fork of CommonTokenStream as explained in https://stackoverflow.com/questions/29060496/allow-whitespace-sections-antlr4/29115489#29115489
class MultiChannelTokenStream(BufferedTokenStream):
    __slots__ = 'channels'
    
    def __init__(self, lexer:Lexer, channels:List[int]=[Token.DEFAULT_CHANNEL]):
        super().__init__(lexer)
        self.channels = channels

    def disable(self, channel:int):
        if channel not in self.channels:
            return
        self.channels.remove(channel)
        self.index = self.nextTokenOnChannel(self.index)

    def enable(self, channel:int):
        if channel in self.channels:
            return
        self.channels += [channel]
        i = self.index
        while i>0 and i == channel:
            i -= 1
        self.index = i

    def adjustSeekIndex(self, i:int):
        return self.nextTokenOnChannel(i)

    def LB(self, k:int):
        if k==0 or (self.index-k)<0:
            return None
        i = self.index
        n = 1
        # find k good tokens looking backwards
        while n <= k:
            # skip off-channel tokens
            i = self.previousTokenOnChannel(i - 1)
            n += 1
        if i < 0:
            return None
        return self.tokens[i]

    def LT(self, k:int):
        self.lazyInit()
        if k == 0:
            return None
        if k < 0:
            return self.LB(-k)
        i = self.index
        n = 1 # we know tokens[pos] is a good one
        # find k good tokens
        while n < k:
            # skip off-channel tokens, but make sure to not look past EOF
            if self.sync(i + 1):
                i = self.nextTokenOnChannel(i + 1)
            n += 1
        return self.tokens[i]

    # Count EOF just once.#/
    def getNumberOfOnChannelTokens(self):
        n = 0
        self.fill()
        for i in range(0, len(self.tokens)):
            t = self.tokens[i]
            if t.channel in self.channels:
                n += 1
            if t.type==Token.EOF:
                break
        return n

    # Given a starting index, return the index of the next token on one of the channels.
    #  Return i if tokens[i] is on one of the channels.  Return the index of the EOF token
    # if there are no tokens on one of the channels between i and EOF.
    #/
    def nextTokenOnChannel(self, i:int):
        self.sync(i)
        if i>=len(self.tokens):
            return len(self.tokens) - 1
        token = self.tokens[i]
        while token.channel not in self.channels:
            if token.type==Token.EOF:
                return i
            i += 1
            self.sync(i)
            token = self.tokens[i]
        return i

    # Given a starting index, return the index of the previous token on channel.
    #  Return i if tokens[i] is on channel. Return -1 if there are no tokens
    #  on channel between i and 0.
    def previousTokenOnChannel(self, i:int):
        while i>=0 and self.tokens[i].channel not in self.channels:
            i -= 1
        return i



import pathlib
import re
from setuptools import find_packages, setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()


def find_version(filename):
    _version_re = re.compile(r'__version__ = "(.*)"')
    for line in open(filename):
        version_match = _version_re.match(line)
        if version_match:
            return version_match.group(1)

VERSION = find_version("ldpy/__init__.py")

PACKAGES = find_packages(exclude=("examples*", "test*"))

# This call to setup() does all the work
setup(
    name="linked-data-python",
    version=VERSION,
    description="The python package \"linked-data-python\" can rewrite a .ldpy file into an equivalent python code.",
    author="Maxime Lefrançois",
    author_email="maxime.lefrancois@emse.fr",
    url="https://gitlab.com/coswot/ldpy",
    license="MIT",
    platforms=["any"],
    python_requires=">=3.7",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Operating System :: OS Independent",
        "Natural Language :: English",
    ],
    long_description=README,
    long_description_content_type="text/markdown",
    packages=PACKAGES,
    include_package_data=True,
    install_requires=["antlr4", "rdflib"],
    entry_points={
        "console_scripts": [
            "ldpy=ldpy.__main__:main",
        ]
    },
)